from abc import abstractmethod
from typing import TypeVar
from typing_extensions import Protocol

from kashmir.core.data import Id, Release
from kashmir.core.entities.project import Project
from kashmir.core.protocols import Entity


class PRepository(Protocol):

    @abstractmethod
    def get(self, id: Id) -> Entity:
        raise NotImplementedError


class PProjectRepository(PRepository):

    @abstractmethod
    def get(self, id: Id) -> Project:
        raise NotImplementedError

    @abstractmethod
    def create_release(self, release: Release):
        raise NotImplementedError



from typing import Optional

import gitlab  # type: ignore

from kashmir.core.data import Version, Release
from kashmir.core.providers import SCMProvider


class GitLab(SCMProvider):

    def __init__(self, server: str, token: str, project_id: str):
        gl = gitlab.Gitlab(server, token)
        self.project = gl.projects.get(project_id)

    @property
    def version(self) -> Optional[Version]:
        """
        This method tries to find the last version for the project.

        In order for this tho work, it's expected that the version match the
        following pattern <<YEAR>>.<<MILESTONE>>.<<FIX>>
        :return: Version
        """
        releases = self.project.releases.list()
        try:
            last_release = releases[0]
            version_components = map(int, last_release.name.split('.'))
            return Version(*version_components)
        except (IndexError, TypeError):
            return None

    def new_release(self, release: Release):
        branch_body = {
            "branch": f"release-{release.version}",
            "ref": "master"
        }
        release_body = {
            "name": str(release.version),
            "tag_name": str(release.version),
            "description": f"Release {release.version}",
            "ref": f"release-{release.version}"
        }
        self.project.branches.create(branch_body)
        self.project.releases.create(release_body)



from abc import abstractmethod
from typing import Any
from typing_extensions import Protocol

from kashmir.core.data import Version, Release, Id


class SCMProvider(Protocol):
    """
    A Source Control Management provider.
    """
    project: Any

    @property
    @abstractmethod
    def version(self):
        raise NotImplementedError

    @abstractmethod
    def new_release(self, release: Release):
        raise NotImplementedError



from dataclasses import dataclass
from datetime import datetime

from kashmir.core.data import Version, Release, Id
from kashmir.core.protocols.entity import Entity


@dataclass
class Project(Entity):

    def __init__(self, id: Id, version: Version):
        self.id = id
        self.version = version

    def new_release(self) -> Release:
        current_year = datetime.now().year
        if self.version and self.version.year == current_year:
            year, release, fix = self.version.year, self.version.release + 1, self.version.fix
        else:
            year, release, fix = current_year, 1, 0
        new_version = Version(year, release, fix)
        return Release(self.id, new_version)

    def new_fix(self) -> Release:
        new_version = Version(self.version.year,
                              self.version.release,
                              self.version.fix + 1)
        return Release(self.id, new_version)

    def __str__(self):
        return f"Project<{self.id}> {self.version}"
